package engine;

import static org.junit.Assert.assertEquals;
import static org.junit.Assert.assertTrue;

import java.util.ArrayList;
import java.util.Arrays;
import java.util.Collections;
import java.util.List;

import org.junit.After;
import org.junit.AfterClass;
import org.junit.Before;
import org.junit.BeforeClass;
import org.junit.Test;

import board.BoardCount;
import board.Board;
import engine.MiniMax.ColumnScore;


public class Tests {

	@BeforeClass
	public static void setUpClass() throws Exception {
		// Code executed before the first test method
	}

	@Before
	public void setUp() throws Exception {
		// Code executed before each test
	}

	@Test
	public void testBoardAdjacencyCount() {
		
		Board board = new BoardCount();

		board.setMove(1, false);
		board.setMove(2, true);
		board.setMove(3, true);
		board.setMove(4, true);
		board.setMove(5, false);
		board.setMove(6, false);
		board.setMove(6, false);
		
		assertEquals(board.countAdjacent(0,false), 1);
		assertEquals(board.countAdjacent(1,false), 1);
		assertEquals(board.countAdjacent(2,false), 1);
		assertEquals(board.countAdjacent(3,false), 0);
		assertEquals(board.countAdjacent(4,false), 1);
		assertEquals(board.countAdjacent(5,false), 3);
		assertEquals(board.countAdjacent(6,false), 1);
		assertEquals(board.countAdjacent(0,true), 0);
		assertEquals(board.countAdjacent(1,true), 1);
		assertEquals(board.countAdjacent(2,true), 2);
		assertEquals(board.countAdjacent(3,true), 3);
		assertEquals(board.countAdjacent(4,true), 2);
		assertEquals(board.countAdjacent(5,true), 1);
		assertEquals(board.countAdjacent(6,true), 0);

	}

	@Test
	public void testBoardThreatAboveThreat() {

		BoardCount board = new BoardCount();
		int n = board.getFeatureVector().length;

		double[] myWeights = new double[n];
		for(int i=0; i<n/2; ++i)
			myWeights[i] = 1;
		for(int i=n/2; i<n; ++i)
			myWeights[i] = 0;
		double[] oppWeights = new double[n];
		for(int i=0; i<n/2; ++i)
			oppWeights[i] = 0;
		for(int i=n/2; i<n; ++i)
			oppWeights[i] = 1;

		board.setMove(0, false);
		board.setMove(1, false);
		board.setMove(2, false);
		assertEquals(9,(int)Sarsa.calculateScore(myWeights, board.getFeatureVector()));
		assertTrue(board.isThreat(false, 3, 0));
	
		board.setMove(0, true);
		board.setMove(1, true);
		board.setMove(2, true);
		assertEquals(4,(int)Sarsa.calculateScore(myWeights, board.getFeatureVector()));

		assertTrue(board.isThreat(false, 3, 0));
		assertTrue(board.isThreat(true, 3, 1));
		assertEquals(9,(int)Sarsa.calculateScore(oppWeights, board.getFeatureVector())); // dont forget that on the second row, there is an extra diagonal counted as a single!
	
	}
	
	@Test
	public void testBoardFeaturesCount() {
		
		Board board = new BoardCount();
		int n = board.getFeatureVector().length;
		
		double[] weights = new double[n];

		board.setMove(0, false);
		board.setMove(0, true);
		board.setMove(0, false);

		for(int i=0; i<n; ++i)
			weights[i] = 1;
		/*for(int i=3; i<Constants.nFeatures-3; ++i)
			weights[i] = 1;
		for(int i=0; i<4; ++i)
			weights[i] = 1;*/
		//assertEquals((int)Sarsa.calculateScore(weights, board.getFeatureVector()), 7);
		
		for(int i=0; i<n; ++i)
			weights[i] = 1;
		assertEquals((int)Sarsa.calculateScore(weights, board.getFeatureVector()), 7);
		
		board.clear();
		board.setMove(3, false);
		
		board.setMove(0, false);
		assertEquals((int)Sarsa.calculateScore(weights, board.getFeatureVector()), 9);
		board.popMove();

		board.setMove(1, false);
		assertEquals((int)Sarsa.calculateScore(weights, board.getFeatureVector()), 9);
		board.popMove();

		board.setMove(2, false);
		assertEquals((int)Sarsa.calculateScore(weights, board.getFeatureVector()), 9);
		board.popMove();

		board.setMove(3, false);
		assertEquals((int)Sarsa.calculateScore(weights, board.getFeatureVector()), 16);
		board.popMove();

		board.setMove(4, false);
		assertEquals((int)Sarsa.calculateScore(weights, board.getFeatureVector()), 11);
		board.popMove();

		board.setMove(5, false);
		assertEquals((int)Sarsa.calculateScore(weights, board.getFeatureVector()), 10);
		board.popMove();

		board.setMove(6, false);
		assertEquals((int)Sarsa.calculateScore(weights, board.getFeatureVector()), 10);
		board.popMove();

		board.setMove(7, false);
		assertEquals((int)Sarsa.calculateScore(weights, board.getFeatureVector()), 10);
		board.popMove();
		
		
		board.clear();
		board.setMove(0, false);
		
		board.setMove(0, false);
		assertEquals((int)Sarsa.calculateScore(weights, board.getFeatureVector()), 6);
		board.popMove();

		board.setMove(1, false);
		assertEquals((int)Sarsa.calculateScore(weights, board.getFeatureVector()), 6);
		board.popMove();

		board.setMove(2, false);
		assertEquals((int)Sarsa.calculateScore(weights, board.getFeatureVector()), 7);
		board.popMove();

		board.setMove(3, false);
		assertEquals((int)Sarsa.calculateScore(weights, board.getFeatureVector()), 9);
		board.popMove();

		board.setMove(4, false);
		assertEquals((int)Sarsa.calculateScore(weights, board.getFeatureVector()), 10);
		board.popMove();

		board.setMove(5, false);
		assertEquals((int)Sarsa.calculateScore(weights, board.getFeatureVector()), 8);
		board.popMove();

		board.setMove(6, false);
		assertEquals((int)Sarsa.calculateScore(weights, board.getFeatureVector()), 7);
		board.popMove();

		board.setMove(7, false);
		assertEquals((int)Sarsa.calculateScore(weights, board.getFeatureVector()), 6);
		board.popMove();
		
	}
	
	@Test
	public void testBoardWinner() {
		
		Board board = new BoardCount();


		board.setMove(1,false);
		board.setMove(2,false);
		board.setMove(3,false);
		board.setMove(1,true);
		board.setMove(2,true);
		board.setMove(3,true);

		assertEquals(board.winner(), -1);
		
		board.setMove(1,false);
		board.setMove(2,false);
		board.setMove(3,false);
		board.setMove(1,true);
		board.setMove(2,true);
		board.setMove(3,true);

		assertEquals(board.winner(), -1);
		
		board.setMove(0,false);
		
		assertEquals(board.winner(), 0);

		
		board.clear();

		board.setMove(1,true);
		board.setMove(2,true);
		board.setMove(3,true);
		board.setMove(1,false);
		board.setMove(2,false);
		board.setMove(3,false);

		assertEquals(board.winner(), -1);
		
		board.setMove(1,true);
		board.setMove(2,true);
		board.setMove(3,true);
		board.setMove(1,false);
		board.setMove(2,false);
		board.setMove(3,false);

		assertEquals(board.winner(), -1);
		
		board.setMove(0,true);
		
		assertEquals(board.winner(), 1);

		board.clear();
		
		
		/*
		 * |o|o|x|o|o|o|x|o|
		 * |x|x|o|x|x|x|o|x|
		 * |o|o|x|o|o|o|x|o|
		 * |x|x|o|x|x|x|o|x|
		 * |o|o|x|o|o|o|x|o|
		 * |x|x|o|x|x|x|o|x|
		 * -----------------
		 */
		
		for (int i=0; i<Constants.nCols; ++i) {
			if (i==2||i==6) continue;
			for(int j=0; j<Constants.nRows; ++j) {
				assertEquals(board.winner(), -1);
				board.setMove(i, j%2==0);
			}
		}
		
		for(int j=0; j<Constants.nRows; ++j) {
			assertEquals(board.winner(), -1);
			board.setMove(2, j%2!=0);
			board.setMove(6, j%2!=0);
		}
		
		assertEquals(board.winner(), 2);

		board.clear();

		/*
		 * | | | | | | | | |
		 * | | | | | | | | |
		 * | | | | |x| | | |
		 * | | | |x|x| | | |
		 * | | |x|x|x| | | |
		 * | |x|x|x|o| | | |
		 * -----------------
		 */
		
		for (int i=1; i<5; ++i) {
			for(int j=0; j<i; ++j) {
				assertEquals(board.winner(), -1);
				board.setMove(i, !(i==4&&j==0));
			}
		}
		
		assertEquals(board.winner(), 1);

		
	}
	
	@Test
	public void testEpsilonDecider() {

		Board board = new BoardCount();
		int n = board.getFeatureVector().length;
		ExploitExplore deciderEpsilon = new DeciderEpsilonGreedy(1,1);
		
		double[] weights = new double[n];
		for(int i=0; i<n; ++i)
			weights[i] = 1;
		
		int nextMove = deciderEpsilon.getNextMove(board, weights, 1);
		assertTrue(nextMove>=2 && nextMove<=3);

		// These results are simply based on the counts in the testBoardFeaturesCount test
		
		board.setMove(3, false);
		assertTrue(deciderEpsilon.getNextMove(board, weights, 1) == 3);

		board.popMove();
		board.setMove(0, false);
		assertTrue(deciderEpsilon.getNextMove(board, weights, 1) == 4);

	}
	
	@Test
	public void testMiniMaxNoHeuristic() {

		Board board = new BoardCount(){
			@Override
			public double[] getFeatureVector() {
				double[] v = super.getFeatureVector();
				int move = this.popMove();

				for(int i=0; i<v.length; ++i)
					v[i] = 0;
				if (move>=0) {
					setMove(move, false);
					v[0] = move;
				}
				return v;
			}
			@Override
			public int countAdjacent(int column, boolean isOpponent) {
				return 0; // initial order, since the sort algorithm guarantees equal elements won't be reordered
			}
		};
		int n = board.getFeatureVector().length;
		MiniMax minimax1 = new MiniMax(1);
		
		double[] weights = new double[n];
		for(int i=0; i<n; ++i)
			weights[i] = 1;
		
		Double[] scores = minimax1.getColumnScores(board, weights);

		for(int i=0; i<scores.length; ++i) {
			assertEquals((int)(double)scores[i], i);
		}
		
		ColumnScore cs = minimax1.getMaxColumnScore(board, weights);
		assertEquals(scores.length-1, (int)(double) cs.getScore());
		assertEquals(scores.length-1, cs.getCol());
	}
	
	@Test
	public void testMiniMaxHeuristic() {

		Board board = new BoardCount(){
			@Override
			public double[] getFeatureVector() {
				double[] v = super.getFeatureVector();
				int move = this.popMove();

				for(int i=0; i<v.length; ++i)
					v[i] = 0;
				if (move>=0) {
					setMove(move, false);
					v[0] = move;
				}
				return v;
			}
			@Override
			public int countAdjacent(int column, boolean isOpponent) {
				return (int)Math.floor(Math.random()*1000);
			}
		};
		int n = board.getFeatureVector().length;
		MiniMax minimax1 = new MiniMax(1);
		
		double[] weights = new double[n];
		for(int i=0; i<n; ++i)
			weights[i] = 1;
		
		Double[] scores;
		
		// a few times, since it's random
		for(int i=0; i<50; ++i) {

			scores = minimax1.getColumnScores(board, weights);
			
			for(int j=0; j<scores.length; ++j) {
				assertEquals((int)(double)scores[j], j);
			}

			ColumnScore cs = minimax1.getMaxColumnScore(board, weights);
			assertEquals(scores.length-1, (int)(double) cs.getScore());
			assertEquals(scores.length-1, cs.getCol());
		}
		
	}
	
	@Test
	public void testMiniMax() {

		Board board = new BoardCount();
		int n = board.getFeatureVector().length;
		MiniMax minimax1 = new MiniMax(1);
		
		double[] weights = new double[n];
		for(int i=0; i<n; ++i)
			weights[i] = 1;
		
		Double[] scores = minimax1.getColumnScores(board, weights);

		// everything*2 since currently we count "twice" once for the delta and once separately
		assertEquals(3, (int)(double)scores[0]);
		assertEquals(4, (int)(double)scores[1]);
		assertEquals(5, (int)(double)scores[2]);
		assertEquals(7, (int)(double)scores[3]);
		assertEquals(7, (int)(double)scores[4]);
		assertEquals(5, (int)(double)scores[5]);
		assertEquals(4, (int)(double)scores[6]);
		assertEquals(3, (int)(double)scores[7]);

		assertEquals(7, (int)(double) minimax1.getMaxColumnScore(board, weights).getScore());

		ColumnScore cs = minimax1.getMaxColumnScore(board, weights);
		assertEquals(7, (int)(double) cs.getScore());
		assertEquals(3, cs.getCol());

		board.setMove(3, false);
		scores = minimax1.getColumnScores(board, weights);

		assertEquals(9, (int)(double)scores[0]);
		assertEquals(9, (int)(double)scores[1]);
		assertEquals(9, (int)(double)scores[2]);
		assertEquals(16, (int)(double)scores[3]);
		assertEquals(11, (int)(double)scores[4]);
		assertEquals(10, (int)(double)scores[5]);
		assertEquals(10, (int)(double)scores[6]);
		assertEquals(10, (int)(double)scores[7]);

		cs = minimax1.getMaxColumnScore(board, weights);
		assertEquals(16, (int)(double) cs.getScore());
		assertEquals(3, cs.getCol());
		
		
		board.clear();
		board.setMove(0, false);
		scores = minimax1.getColumnScores(board, weights);

		assertEquals(6, (int)(double)scores[0]);
		assertEquals(6, (int)(double)scores[1]);
		assertEquals(7, (int)(double)scores[2]);
		assertEquals(9, (int)(double)scores[3]);
		assertEquals(10, (int)(double)scores[4]);
		assertEquals(8, (int)(double)scores[5]);
		assertEquals(7, (int)(double)scores[6]);
		assertEquals(6, (int)(double)scores[7]);

		cs = minimax1.getMaxColumnScore(board, weights);
		assertEquals(10, (int)(double) cs.getScore());
		assertEquals(4, cs.getCol());
		
	}

	@Test
	public void testMiniMax1() {

		Board board = new TamperedBoard();
		int n = board.getFeatureVector().length;
		
		MiniMax minimax1 = new MiniMax(4);
		
		double[] weights = new double[n];
		for(int i=0; i<n; ++i)
			weights[i] = 1;
		
		Double[] scores = minimax1.getColumnScores(board, weights);

		assertEquals(4, (int)(double)scores[0]);
		assertEquals(6, (int)(double)scores[1]);
		assertEquals(7, (int)(double)scores[2]);
		for(int j=3; j<scores.length; ++j)
			assertEquals(null,scores[j]);

		System.out.println("---------------");
		
		ColumnScore cs = minimax1.getMaxColumnScore(board, weights);
		assertEquals(7, (int)(double) cs.getScore());
		assertEquals(2, cs.getCol());

		System.out.println("Expected above (below -------------)");
		System.out.println("Prune: 8 >= 6");
		System.out.println("Prune: 4 >= 3");
		System.out.println();
	
	}
	
	@Test
	public void testMiniMax2() {

		Board board = new TamperedBoard(9);
		int n = board.getFeatureVector().length;
		
		MiniMax minimax1 = new MiniMax(4);
		
		double[] weights = new double[n];
		for(int i=0; i<n; ++i)
			weights[i] = 1;
		
		Double[] scores = minimax1.getColumnScores(board, weights);
		
		assertEquals(4, (int)(double)scores[0]);
		assertEquals(8, (int)(double)scores[1]);
		assertEquals(7, (int)(double)scores[2]);
		for(int j=3; j<scores.length; ++j)
			assertEquals(null,scores[j]);

		System.out.println("---------------");

		ColumnScore cs = minimax1.getMaxColumnScore(board, weights);
		assertEquals(8, (int)(double) cs.getScore());
		assertEquals(1, cs.getCol());
		
		System.out.println("Expected above (below -------------)");
		System.out.println("Prune: 8 >= 6");
		System.out.println("Prune: 4 >= 3");
		System.out.println("Prune: 9 >= 8");
		System.out.println();
	
	}
	
	@Test
	public void testMiniMax3() {

		Board board = new TamperedBoard(){
			 @Override
			public int countAdjacent(int column, boolean isOpponent) {
				return column; // reverse order
			}
		};
		int n = board.getFeatureVector().length;

		MiniMax minimax1 = new MiniMax(4);
		
		double[] weights = new double[n];
		for(int i=0; i<n; ++i)
			weights[i] = 1;

		Double[] scores = minimax1.getColumnScores(board, weights);
		
		assertEquals(4, (int)(double)scores[0]);
		assertEquals(6, (int)(double)scores[1]);
		assertEquals(7, (int)(double)scores[2]);
		for(int j=3; j<scores.length; ++j)
			assertEquals(null,scores[j]);

		System.out.println("---------------");
		
		ColumnScore cs = minimax1.getMaxColumnScore(board, weights);
		assertEquals(7, (int)(double) cs.getScore());
		assertEquals(2, cs.getCol());
		
		System.out.println("Expected above (below -------------)");
		System.out.println("Prune: 7 >= 6");
		System.out.println("Prune: 7 >= -1");
		System.out.println("Prune: 7 >= 7");
		System.out.println("Prune: 7 >= 7");
		System.out.println();
	
	}

	@After
	public void tearDown() throws Exception {
		// Code executed after each test
	}

	@AfterClass
	public static void tearDownClass() throws Exception {
		// Code executed after the last test method
	}
	
	private class TamperedBoard extends BoardCount {
		
		private int V = 5;
		
		public TamperedBoard() {
		}
		
		public TamperedBoard(int V) {
			this.V=V;
		}
		
		/**
		 * Takes two lists and returns true if the first is the head of the second
		 * @return
		 */
		public boolean begins(int[] head, List<Integer> list) {
			List<Integer> headList = new ArrayList<Integer>();
		    for (int index = 0; index < head.length; index++)
		    	headList.add(head[index]);
			return list.size()>=head.length && list.subList(0, head.length).equals(headList);
		}
		
		List<Integer> getPath() {
			List<Integer> path = new ArrayList<Integer>();
			int move;
			boolean isOpponent = false;
			
			while((move = this.popMove()) != -1) {
				path.add(move);
			}
			Collections.reverse(path);

			for(int col : path) {
				setMove(col, isOpponent);
				isOpponent = !isOpponent;
			}
			return path;
		}
		
		@Override
		public double[] getFeatureVector() {
			double[] v = super.getFeatureVector();
			
			List<Integer> path = getPath();
			// now we have the path!
			
			for(int i=0; i<v.length; ++i)
				v[i] = 0;

			if (begins(new int[]{0,0}, path))
				v[0] = 4;
			else if (begins(new int[]{0,1}, path))
				v[0] = 7;
			else if (begins(new int[]{1,0,0}, path))
				v[0] = 8;
			else if (begins(new int[]{1,0,1,0}, path))
				v[0] = 6;
			else if (begins(new int[]{1,0,1,1}, path))
				v[0] = 12;
			else if (begins(new int[]{1,1,0,0}, path))
				v[0] = 3;
			else if (begins(new int[]{1,1,0,1}, path))
				v[0] = 6;
			else if (begins(new int[]{1,1,0,2}, path))
				v[0] = -1;
			else if (begins(new int[]{1,1,1}, path))
				v[0] = V;
			else if (begins(new int[]{1,1,2,0}, path))
				v[0] = 6;
			else if (begins(new int[]{1,1,2,1}, path))
				v[0] = 12;
			else if (begins(new int[]{2}, path))
				v[0] = 7;
			
			return v;
		}
		@Override
		public int countAdjacent(int column, boolean isOpponent) {
			return 0; // initial order
		}
		@Override
		public boolean isLegalMove(int move) {
			List<Integer> path = getPath();

			String[] paths = {"00","01","100","1010","1011","1100","1101","1102","111","1120","1121","2"};
			Arrays.sort(paths);
			
			String pathStr="";
			for(int p : path)
				pathStr += p;
			pathStr += move;
			
			for(String p : paths) {
				if (p.startsWith(pathStr))
					return true;
			}
			
			return false;
			
		}
		// we need to override winner, since we override isLegalMove to cap the tree growth, so we need
		// to make sure the winner() reflects that in a position with no legal moves, we consider winner!=-1,
		// and evaluate the feature vector there (otherwise we create nCols children with all null beta values,
		// and propogate rubbish up the tree)
		@Override
		public int winner() {
			int winner = super.winner();
			if (winner==-1) {
				boolean validMoveExists = false;
				for(int i=0;i<Constants.nCols;++i) {
					if (this.isLegalMove(i)) {
						validMoveExists = true;
						break;
					}
				}
				if (!validMoveExists)
					winner=2;
			}
			return winner;
		}
	};
	
}