package guiAgent;

import java.util.Arrays;

import org.eclipse.swt.SWT;
import org.eclipse.swt.events.KeyAdapter;
import org.eclipse.swt.events.KeyEvent;
import org.eclipse.swt.events.MouseAdapter;
import org.eclipse.swt.events.MouseEvent;
import org.eclipse.swt.events.MouseTrackAdapter;
import org.eclipse.swt.graphics.Image;
import org.eclipse.swt.graphics.Rectangle;
import org.eclipse.swt.layout.FillLayout;
import org.eclipse.swt.layout.GridLayout;
import org.eclipse.swt.layout.RowLayout;
import org.eclipse.swt.widgets.Composite;
import org.eclipse.swt.widgets.Display;
import org.eclipse.swt.widgets.Label;
import org.eclipse.swt.widgets.Shell;


public class Visualiser {

	private final int nCols = 8;
	private final int nRows = 6;

	//final Image empty = new Image(display, ClassLoader.getSystemResourceAsStream("resources\\empty.png"));
	private final Image empty;
	private final Image[] pieces, shadows;
	
	private Label[][] holders = new Label[nCols][nRows];

	private int player=1;
	
	public static void main(String[] args) {
		new Visualiser();
	}
	
	public Visualiser() {
				
		final Display display = new Display();
		final Shell shell = new Shell(display);
		
		shell.setText("Connect Four");
	    //shell.setBackground(display.getSystemColor(SWT.COLOR_BLUE));
	    //shell.setBackgroundMode(SWT.INHERIT_FORCE);
	    shell.setSize(1000, 700);

	    // center on screen
		Rectangle screen = display.getPrimaryMonitor().getBounds(), window = shell.getBounds();
		shell.setLocation(screen.x + (screen.width - window.width)/2, screen.y + (screen.height - window.height)/2);
		
		shell.setLayout(new FillLayout());
		
		final Composite grid = new Composite(shell, SWT.NONE);
		GridLayout gridLayout = new GridLayout(nCols, true);
		gridLayout.horizontalSpacing = gridLayout.verticalSpacing = 0;
		grid.setLayout(gridLayout);

		empty = new Image(display, "resources\\empty.png");
		pieces = new Image[]{new Image(display, "resources\\red.png"), new Image(display, "resources\\yellow.png") };
		shadows = new Image[]{new Image(display, "resources\\red_shadow.png"), new Image(display, "resources\\yellow_shadow.png") };
		
		// draw grid
		for(int i=0; i<nCols; ++i) {
			final Composite column = new Composite(grid, SWT.NONE);
			RowLayout rowLayout = new RowLayout(SWT.VERTICAL);
			rowLayout.spacing = rowLayout.marginTop = rowLayout.marginLeft = rowLayout.marginRight = rowLayout.marginBottom = 0;
			column.setLayout(rowLayout);
			
			final int _i = i;
			for(int j=0; j<nRows; ++j) {
				holders[i][j] = new Label(column, SWT.NONE);
				holders[i][j].setImage(empty);
				// would prefer the composite, but the event doesn't bubble in SWT
				holders[i][j].addMouseTrackListener(new MouseTrackAdapter() {
					@Override
					public void mouseEnter(MouseEvent e) {
						switchFirstThatEquals(_i, new Image[]{empty}, shadows[player]);
					}
					@Override
					public void mouseExit(MouseEvent e) {
						switchFirstThatEquals(_i, shadows, empty);
					}
				});
				holders[i][j].addMouseListener(new MouseAdapter() {
					@Override
					public void mouseDown(MouseEvent e) {
						switchFirstThatEquals(_i, shadows, pieces[player]);
						switchFirstThatEquals(_i, new Image[]{empty}, shadows[player]);
					}
				});
			}
		}
		
		shell.addKeyListener(new KeyAdapter() {
			@Override
			public void keyReleased(KeyEvent e) {
				player=1-player;
				for(int i=0; i<nCols && !switchFirstThatEquals(i, shadows, shadows[player]); ++i);
			}
		});

		shell.open();

		while (!shell.isDisposed()) {
			if (!display.readAndDispatch())
				display.sleep();
		}
		display.dispose();

	}

	/**
	 * @return true iff it found and replaced and image
	 */
	private boolean switchFirstThatEquals(int col, Image[] compare, Image replaceWith) {
		int j;
		for(j=nRows-1; j>=0 && !Arrays.asList(compare).contains(holders[col][j].getImage()); --j);
		if (j>=0) { // not full
			holders[col][j].setImage(replaceWith);
			return true;
		}
		return false;
	}

}